#!/usr/bin/python

CONFIG_FILE="config.db"

DEFAULT_HOST="127.0.0.1"
DEFAULT_PORT="8086"
DEFAULT_BASEDIR="/srv/www/webcpl/hosts/"
DEFAULT_PHP_SOCK="unix:/var/run/php5-fpm.sock"
DEFAULT_ADMINPASS="admin"

SSLMODE_HTTP_ONLY = 0
SSLMODE_BOTH = 1
SSLMODE_HTTPS_ONLY = 2

SSLMODE_STRINGS = {
	SSLMODE_HTTP_ONLY: "HTTP Only",
	SSLMODE_BOTH: "HTTP and HTTPS",
	SSLMODE_HTTPS_ONLY: "HTTPS Only"
}


#

import bottle
import cgi
import hashlib
import os
import stat
import sqlite3
import sys
import tempfile
import time
import urllib
import uuid

# Monkeypatch bottle.py to disable DNS lookups (ref: http://blog.est.im/post/34288214582)
__import__('BaseHTTPServer').BaseHTTPRequestHandler.address_string = lambda x:x.client_address[0]

#

class AppDB():
	def __init__(self, path):
		self.sql = sqlite3.connect(path, isolation_level=None); # Autocommit
		self.sql.row_factory = sqlite3.Row # Request dicts instead of lists
		self.createTables()

	def createTables(self):
		self.sql.execute('''
			CREATE TABLE IF NOT EXISTS config (
				k TEXT,
				v TEXT
			)
		''');
		self.sql.execute('''
			CREATE UNIQUE INDEX IF NOT EXISTS config_k ON config (k)
		''');
		self.sql.execute('''
			CREATE TABLE IF NOT EXISTS sites (
				siteid TEXT PRIMARY KEY,
				description TEXT,
				userdir TEXT,
				allowPhp INTEGER,
				rewriteToIndex INTEGER
			)
		''');
		self.sql.execute('''
			CREATE TABLE IF NOT EXISTS routes (
				id INTEGER PRIMARY KEY AUTOINCREMENT,
				siteid TEXT,
				hostname TEXT,
				redirect_url TEXT,
				sslMode INTEGER,
				sslid INTEGER
			)
		''');
		self.sql.execute('''
			CREATE TABLE IF NOT EXISTS sslcerts (
				id INTEGER PRIMARY KEY AUTOINCREMENT,
				description STRING,
				ssl_crt TEXT,
				ssl_chain TEXT,
				ssl_key TEXT,
				cache_expiry INTEGER
			)
		''');
		self.sql.execute('''
			CREATE TABLE IF NOT EXISTS interfaces (
				id INTEGER PRIMARY KEY AUTOINCREMENT,
				ipaddr TEXT,
				port_http INTEGER,
				port_https INTEGER,
				addByDefault INTEGER
			)
		''');
		self.sql.execute('''
			CREATE TABLE IF NOT EXISTS route_interfaces (
				id INTEGER PRIMARY KEY AUTOINCREMENT,
				routeid INTEGER,
				interfaceid INTEGER
			)
		''');
		self.sql.execute('''
			CREATE INDEX IF NOT EXISTS route_interfaces_routeid ON route_interfaces (routeid)
		''');
		
		current_schema_version = 100;
		if int(self.getConfig('db.schema', current_schema_version)) < current_schema_version:
			pass
		
	def getConfig(self, key, default=None):
		ret = self.sql.execute("SELECT v FROM config WHERE k = ?", [key])
		rows = ret.fetchall()
		if len(rows) == 1:
			return rows[0][0]
		else:
			sys.stderr.write("[WARN] Missing key '%s' in config DB, saving default '%s'\n" % (key, default))
			self.sql.execute("INSERT INTO config (k, v) VALUES (?, ?)", [key, default]);
			return default
	
	def getSites(self):
		return self.sql.execute("SELECT * FROM sites").fetchall()
	
	def getRoutes(self):
		return self.sql.execute('SELECT * FROM routes').fetchall()
		
	def getCerts(self):
		return self.sql.execute('SELECT * FROM sslcerts').fetchall()
		
	def getInterfaces(self):
		return self.sql.execute('SELECT * FROM interfaces').fetchall()
	
	def getSite(self, siteid):
		ret = self.sql.execute("SELECT * FROM sites WHERE siteid = ?", [siteid]).fetchall()
		return ret[0] if len(ret) == 1 else False;
	
	def getRoute(self, routeid):
		ret = self.sql.execute("SELECT * FROM routes WHERE id = ?", [routeid]).fetchall()
		return ret[0] if len(ret) == 1 else False;
		
	def getCert(self, sslid):
		ret = self.sql.execute("SELECT * FROM sslcerts WHERE id = ?", [sslid]).fetchall()
		return ret[0] if len(ret) == 1 else False;
		
	def getInterface(self, id):
		ret = self.sql.execute("SELECT * FROM interfaces WHERE id = ?", [id]).fetchall()
		return ret[0] if len(ret) == 1 else False;
		
	def getInterfaceIDsForRoute(self, routeid):
		dbh = self.sql.execute("SELECT * FROM route_interfaces WHERE routeid = ?", [routeid]).fetchall()
		return [x["interfaceid"] for x in dbh];
		
	def insertRouteRow(self, props):
		return self._insert("routes", props);
		
	def updateRouteRow(self, id, props):
		return self._update("routes", props, "id = ?", [id])
		
	def deleteRouteRow(self, id):
		self.sql.execute("DELETE FROM routes WHERE id = ?", [id])
		self.sql.execute("DELETE FROM route_interfaces WHERE routeid = ?", [id])
		
	def insertSiteRow(self, props):
		return self._insert("sites", props);
		
	def updateSiteRow(self, siteid, props):
		return self._update("sites", props, "siteid = ?", [siteid])
		
	def deleteSiteRow(self, siteid):
		self.sql.execute("DELETE FROM sites WHERE siteid = ?", [siteid])
		
	def insertInterfaceRow(self, props):
		return self._insert("interfaces", props);
		
	def updateInterfaceRow(self, id, props):
		return self._update("interfaces", props, "id = ?", [id])
		
	def deleteInterfaceRow(self, id):
		self.sql.execute("DELETE FROM interfaces WHERE id = ?", [id])
		self.sql.execute("DELETE FROM route_interfaces WHERE interfaceid = ?", [id])
		
	def insertCertRow(self, props):
		return self._insert("sslcerts", props)
		
	def updateCertRow(self, id, props):
		return self._update("sslcerts", props, "id = ?", [id])
		
	def deleteCertRow(self, id):
		self.sql.execute("DELETE FROM sslcerts WHERE id = ?", [id])
		self.sql.execute("UPDATE routes SET sslid = ? WHERE sslid = ?", [0, id])
		self.sql.execute("UPDATE routes SET sslmode = ? WHERE sslid = ? AND sslmode IN (?, ?)", [SSLMODE_HTTP_ONLY, id, SSLMODE_BOTH, SSLMODE_HTTPS_ONLY])
		
	def _insert(self, table, props):
		cursor = self.sql.cursor()
		cursor.execute(
			"INSERT INTO "+table+"("+ (",".join(props.keys()))+") VALUES (" + (",".join( ["?"]*len(props.keys()) )) + ")",
			props.values()
		)
		return cursor.lastrowid
	
	def _update(self, table, props, whereClause, whereProps):
		self.sql.execute(
			"UPDATE "+table+" SET " + (",".join(["`"+x+"` = ?" for x in props.keys()])) + " WHERE (" + whereClause + ")",
			props.values() + whereProps
		)
		return True
	
	def dbFold(self, data, key):
		ret = {}
		for row in data:
			ret[ str(row[key]) ] = row
		return ret

#

class AppServer(AppDB):
	def __init__(self, path):
		AppDB.__init__(self, path) # call parent ctor
		
	def getWebProps(self):
		return {
			'host': self.getConfig('web.host', DEFAULT_HOST),
			'port': self.getConfig('web.port', DEFAULT_PORT)
		};
		
	def createNewSite(self, options):
		basedir = self.getConfig('basedir', DEFAULT_BASEDIR)
		siteid = str(uuid.uuid4())
		userdir = basedir + siteid + "/"
		os.mkdir(userdir)
		os.mkdir(userdir + "wwwroot/")
		
		options["siteid"] = siteid
		options["userdir"] = userdir
		
		self.insertSiteRow(options)
		return siteid
		
	def createNewCert(self, options):
		options["cache_expiry"] = self.getSSLExpiryTimestamp( options["ssl_crt"] )
		return self.insertCertRow(options)
		
	def updateExistingCert(self, id, options):
		options["cache_expiry"] = self.getSSLExpiryTimestamp( options["ssl_crt"] )
		return self.updateCertRow(id, options)
		
	def getCurrentIPAddresses(self):
		return os.popen("ip addr|grep inet|cut -d' ' -f6").read().split("\n")[:-1]
		
	def hashPassword(self, pass_str):
		return hashlib.sha256(pass_str).hexdigest()
		
	def isAdminPassword(self, pass_str):
		hashed = self.hashPassword(pass_str)
		expect_hash = self.getConfig("webui.adminhash", self.hashPassword(DEFAULT_ADMINPASS))
		return hashed == expect_hash
		
	def getSSLExpiryTimestamp(self, crtbody):
		tmpfh, tmpf = tempfile.mkstemp()
		os.write(tmpfh, crtbody)
		ret = os.popen('date --date="$(cat '+tmpf+' | openssl x509 -noout -enddate | cut -f2 -d=)" "+%s"').read()
		os.unlink( tmpf )
		return int(ret)
		
	def genConfig(self):
		ret = "# Autogenerated by webcpl.py\n\n";
		siteLookup = self.dbFold(self.getSites(), "siteid");
		sslLookup = self.dbFold(self.getCerts(), "id");
		interfaceLookup = self.dbFold(self.getInterfaces(), "id");
		
		basedir = self.getConfig('basedir', DEFAULT_BASEDIR)		
		phpsock = self.getConfig('nginx.phpsock', DEFAULT_PHP_SOCK)
		
		enable_php = '''
	# Enable php
	location ~ \.php$ {
		try_files $uri =404;
		fastcgi_pass   ''' + phpsock + ''';
		fastcgi_index  index.php;
		fastcgi_param  SCRIPT_FILENAME $document_root$fastcgi_script_name;
		include fastcgi_params;
	}
		''' + "\n";
		
		# Build all routes
		
		for route in self.getRoutes():
			ret += "server { \n"
			
			for interfaceid in self.getInterfaceIDsForRoute(route["id"]):
				interface = interfaceLookup[interfaceid]
				if route["sslMode"] != SSLMODE_HTTPS_ONLY:
					ret += "\t listen %s %s ; \n" % (interface["ipaddr"], interface["port_http"])
				if route["sslMode"] != SSLMODE_HTTP_ONLY:
					ret += "\t listen %s %s ; \n" % (interface["ipaddr"], interface["port_https"])
				# To enforce HTTPS, create a separate route HTTPS_ONLY with redirect_url set
				
			if route["sslid"] != SSLMODE_HTTP_ONLY:
				ret += "\t #sslkeyfile %s ;\n" % (basedir+"ssl/"+str(route["sslid"])+".key")
				ret += "\t #sslcrtfile %s ;\n" % (basedir+"ssl/"+str(route["sslid"])+".crt")
			
			ret += "\t server_name %s; \n" % route['hostname']
	
			if (len(route["redirect_url"])):
				ret += "\t location / {\n"
				ret += "\t\t redirect %s break; \n" % route['redirect_url'];
				ret += "\t } \n"
				
			else:
				site = siteLookup[ route["siteid"] ]
				index = ["index.htm", "index.html"]
				
				if site["allowPhp"] == 1:
					index.append("index.php")
					ret += enable_php
				
				ret += "\t index %s ; \n" % " ".join(index)
				
				ret += "\t root %s ; \n" % (site["userdir"]+"wwwroot/")
				
				if site["rewriteToIndex"] == 1:
					ret += "\t location / { \n"
					ret += "\t\t try_files $1 index.php ;\n"
					ret += "\t } \n"
				
			ret += "} \n"
		
		# Add 444/drop as fallback default for all non-https interfaces
				
		ret += "server { \n"
		for interface in interfaceLookup.values():
			print interface
			ret += "\t listen %s:%s default_server ;\n" % ( interface["ipaddr"], str(interface["port_http"]) )
		ret += "\t server_name _ ;\n"
		ret += "\t return 444 ;\n"
		ret += "} \n"
		
		return ret;
#

def HTML_bool(name, val, yes_str="Yes", no_str="No"):
	return '''
		<label>
			<input type="radio" name="'''+cgi.escape(name)+'''" value="1" '''+("" if val == 0 else "checked")+'''>
			'''+cgi.escape(yes_str)+'''
		</label>
		&nbsp;
		<label>
			<input type="radio" name="'''+cgi.escape(name)+'''" value="0" '''+("checked" if val == 0 else "")+'''>
			'''+cgi.escape(no_str)+'''
		</label>
	''';

def HTML_multiOptions(vals, selopts):
	ret = ""
	for key in selopts.keys():
		ret += '<option value="' + cgi.escape(str(key)) + '" '+("selected" if key in vals else "")+'>'+cgi.escape(selopts[key])+'</option>'
	return ret
	
def HTML_select(name, val, selopts):
	return "<select name=\"" + name + "\">" + HTML_multiOptions([val], selopts) + "</select>"
	
def HTML_multi(name, vals=[], selopts={}):
	print vals
	print selopts
	return '<select name="' + name +'" multiple>' + HTML_multiOptions(vals, selopts) + '</select>'
	
	
#

class AppWebUI(object):
	def __init__(self, srv):
		self.srv = srv
		self.sessions = {}
		
		# Configure routes
		bottle.route('/', method='GET', callback=self.routeIndex)
		bottle.route('/login', method='GET', callback=self.routeLogin)
		bottle.route('/sites', method='GET', callback=self.routeSites)
		bottle.route('/site/<siteid>', method='GET', callback=self.routeSite)
		bottle.route('/routes', method='GET', callback=self.routeRoutes)
		bottle.route('/route/<id>', method='GET', callback=self.routeRoute)
		bottle.route('/sslcerts', method='GET', callback=self.routeSSLCerts)
		bottle.route('/sslcert/<id>', method='GET', callback=self.routeSSLCert)
		bottle.route('/interfaces', method='GET', callback=self.routeInterfaces)
		bottle.route('/interface/<id>', method='GET', callback=self.routeInterface)
		bottle.route('/genconfig', method='GET', callback=self.routeGenConfig)
		bottle.route('/error/<message>', method='GET', callback=self.routeError)
		
		bottle.route('/post/login', method='POST', callback=self.processLogin)
		bottle.route('/post/logout', method='POST', callback=self.processLogout)
		bottle.route('/post/updateSite', method='POST', callback=self.processUpdateSite)
		bottle.route('/post/updateRoute', method='POST', callback=self.processUpdateRoute)
		bottle.route('/post/updateCert', method='POST', callback=self.processUpdateCert)
		bottle.route('/post/updateInterface', method='POST', callback=self.processUpdateInterface)
		
	def run(self):
		bottle.run(**srv.getWebProps())
		
		# global
	@bottle.hook("before_request")
	def preroute():
		bottle.response.set_header('Server', 'webcpl/1.0')
		
	def sessionID(self):
		sessionid = bottle.request.get_cookie('SESSIONID', str(uuid.uuid4()))
		bottle.response.set_cookie('SESSIONID', sessionid)
		if not sessionid in self.sessions:
			self.sessions[sessionid] = {}
		return sessionid
		
	def isLoggedIn(self):
		session = self.sessions[self.sessionID()]
		return 'isLoggedIn' in session
		
	def redirect(self, url):
		bottle.response.status = 302
		bottle.response.set_header('Location', url)
		return "Redirecting..."
		
	def redirectError(self, errorString):
		return self.redirect("/error/" + urllib.quote(errorString))
		
	def processLogin(self):
		if not "pass" in bottle.request.forms:
			return self.redirectError("Invalid request.");
			
		self.sessions[self.sessionID()].pop("isLoggedIn", None)
		
		if self.srv.isAdminPassword( bottle.request.forms["pass"] ):
			self.sessions[self.sessionID()]["isLoggedIn"] = True
			return self.routeIndex()
		else:
			return self.redirectError("Invalid password.")
		
	def processLogout(self):
		self.sessions[self.sessionID()].pop("isLoggedIn");
		return self.redirectError("Logged out.")
		
	def processGeneric(self, idKey, otherKeys, fnCreate, fnUpdate, fnDelete, redirPrefix):
		if not self.isLoggedIn():
			return self.routeIndex()
		
		if not idKey in bottle.request.forms:
			return self.redirectError("Invalid request.");
		id = bottle.request.forms[idKey];
		
		if "_delete" in bottle.request.forms and bottle.request.forms["_delete"] == "1":
			fnDelete( id )
			return self.redirectError("Deleted.")
		
		else:
			for requiredKey in otherKeys:
				if not requiredKey in bottle.request.forms:
					return self.redirectError("Invalid request.");
			
			filtered = {x : bottle.request.forms[x] for x in otherKeys};
			
			if id == "new":
				id = fnCreate( filtered )
			else:
				fnUpdate( id, filtered )
		
			return self.redirect(redirPrefix + urllib.quote(str(id)))
		
	def processUpdateSite(self):
		return self.processGeneric("siteid", ["description", "allowPhp", "rewriteToIndex"], self.srv.createNewSite, self.srv.updateSiteRow, self.srv.deleteSiteRow, "/site/")
		
	def processUpdateRoute(self):
		return self.processGeneric("id", ["hostname", "siteid", "redirect_url", "sslMode", "sslid"], self.srv.insertRouteRow, self.srv.updateRouteRow, self.srv.deleteRouteRow, "/route/")
		
	def processUpdateInterface(self):
		return self.processGeneric("id", ["ipaddr", "port_http", "port_https"], self.srv.insertInterfaceRow, self.srv.updateInterfaceRow, self.srv.deleteInterfaceRow, "/interface/")
		
	def processUpdateCert(self):
		return self.processGeneric("id", ["description", "ssl_crt", "ssl_chain", "ssl_key"], self.srv.createNewCert, self.srv.updateExistingCert, self.srv.deleteCertRow, "/sslcert/")
		
	def templateBody(self, body):
		ret = '''<!DOCTYPE html>
<html>
	<head>
		<meta charset="UTF-8">
		<title>webcpl</title>
		<style type="text/css">
html, body {
	margin:0; border:0; padding:0;
	width:100%;
	height:100%;
	min-height:100%;
	font-family:"Segoe UI",Meiryo,Helvetica,sans-serif;
	font-size:16px;
}
body {
	position:relative;
}
.padder {
	padding:8px;
}
.menu {
	position:absolute;
	left:0;
	width:200px;
	top:0;
	bottom:0;
	
	background:#333;
	color:white;
}
.menu strong a {
	color:white;
	text-decoration:none;
}
.menu ul {
	padding:0;
}
.menu ul a {
	color:#CCC;
	transition:0.2s linear color;
	text-decoration:none;
}
.menu ul a:hover {
	color:#FFF;
}
.menu form {
	margin-top:1.0em;
}
.content {
	position:absolute;
	overflow-y:scroll;
	left:200px;
	top:0;
	right:0;
	bottom:0;
	
	background:#F8F8F8;
	color:#333;
}
.content a {
	color: black;
}
.content table {
	width:100%;
	margin:1em 0;
	border:1px solid #F0F0F0;
	border-collapse:collapse;
}
.content table td {
	padding:3px;
}
.content table thead tr {
	font-weight:bold;
	background:#F0F0F0;
}
.content table tbody tr:hover td {
	background:lightyellow;
}
.content table tfoot tr {
	font-style:italic;
	background:#F0F0F0;
	text-align:right;
	color:#666;
}
.content textarea {
	width:100%;
	min-height:250px;
}
.content textarea[readonly] {
	border:0;
	background:#F8F8F8;
}
		</style>
	</head>
	<body>
		<div class="menu">
			<div class="padder">
				<strong><a href="/">webcpl</a></strong>
				<ul>
	''';
		if self.isLoggedIn():
			ret += '''
					<li><a href="/sites">Sites</a></li>
					<li><a href="/routes">Routes</a></li>
					<li><a href="/sslcerts">SSL Certificates</a></li>
					<li><a href="/interfaces">Interfaces</a></li>
					<li><a href="/genconfig">Preview config</a></li>
					<li>
						<form method="POST" action="/post/logout">
							<input type="submit" value="Log out">
						</form>
					</li>
			''';
		else:
			ret += '''
					<li><a href="/login">Log in</a></li>
			''';
		
		ret += '''
				</ul>
			</div>
		</div>
		<div class="content">
			<div class="padder">
<!-- {{{ --> 
''' + body + '''
<!-- }}} --> 
			<div class="padder">
		</div>
	</body>
</html>
		''';
		return ret;
		
	def routeLogin(self):
		if self.isLoggedIn(): return self.routeIndex()
	
		return self.templateBody('''
<strong>Login</strong><br>
<p></p>
<form method="POST" action="/post/login">
	<input type="password" name="pass" placeholder="Password" autofocus>
</form>
		''');
		
	def routeSites(self):
		if not self.isLoggedIn(): return self.routeIndex()
		
		allsites = self.srv.getSites()
		html = '''
<strong>Sites</strong>
<p>
	<a href="/site/new">Create new &raquo;</a>
</p>
<table>
	<thead>
		<tr>
			<td>Description</td>
		</tr>
	</thead>
	<tbody>
		''';
		
		for site in allsites:
			html += '''
		<tr>
			<td>
				<a href="/site/''' + cgi.escape(urllib.quote(site["siteid"])) + "\">" + cgi.escape(site["description"]) + '''</a>
			</td>
		</tr>
			''';
		
		html += '''
	</tbody>
	<tfoot>
		<tr>
			<td colspan="1">
				'''+str(len(allsites))+''' site(s)
			</td>
		</tr>
	</tfoot>
</table>
		''';
		
		return self.templateBody(html);
		
	def routeSite(self, siteid):
		if not self.isLoggedIn(): return self.routeIndex()
		
		if siteid == "new":
			site = {
				"description": "New site",
				"userdir": "{autogenerate}/",
				"allowPhp":1,
				"rewriteToIndex":0
			}
		else:
			site = self.srv.getSite(siteid)
			if site == False:
				return self.routeError("Invalid site selected.")
		
		html = '''
<strong>Sites &rsaquo; '''+cgi.escape(site["description"])+'''</strong><br>
<p>
</p>
<form method="POST" action="/post/updateSite">
	<input type="hidden" name="siteid" value="'''+cgi.escape(siteid)+'''">
	<table>
		<thead>
			<tr>
				<td>&nbsp;</td>
				<td>&nbsp;</td>
			</tr>
		</thead>
		<tbody>
			<tr>
				<td>Site ID</td>
				<td><tt>'''+cgi.escape(siteid)+'''</tt></td>
			</tr>
			<tr>
				<td>Description</td>
				<td><input type="text" name="description" value="'''+cgi.escape(site["description"])+'''"></td>
			</tr>
			<tr>
				<td>Home directory</td>
				<td><tt>'''+cgi.escape(site["userdir"])+'''</tt></td>
			</tr>
			<tr>
				<td>Webroot directory</td>
				<td><tt>'''+cgi.escape(site["userdir"])+'''wwwroot/</tt></td>
			</tr>
			<tr>
				<td>Allow PHP</td>
				<td>'''+HTML_bool("allowPhp", site["allowPhp"])+'''</td>
			</tr>
			<tr>
				<td>Rewrite to index.php</td>
				<td>'''+HTML_bool("rewriteToIndex", site["rewriteToIndex"])+'''</td>
			</tr>
		</tbody>
		<tfoot>
			<tr>
				<td>&nbsp;</td>
				<td><input type="submit" value="Save &raquo;"></td>
			</tr>
		</tfoot>
	</table>
</form>
		''';
		
		if siteid != "new":
			html += '''
<form method="POST" action="/post/updateSite">
	<input type="hidden" name="siteid" value="'''+cgi.escape(siteid)+'''">
	<input type="hidden" name="_delete" value="1">
	<input type="submit" value="Delete &raquo;">
</form>
		''';
		
		return self.templateBody(html)
		
	def routeRoutes(self):
		if not self.isLoggedIn(): return self.routeIndex()
		
		routes = self.srv.getRoutes()
		sites = self.srv.dbFold( self.srv.getSites(), "siteid" )
		sslcerts = self.srv.dbFold( self.srv.getCerts(), "id" )
		
		html = '''
<strong>Routes</strong>
<p>
	<a href="/route/new">Create new &raquo;</a>
</p>
<table>
	<thead>
		<tr>
			<td>Hostname</td>
			<td>Connection</td>
			<td>Interfaces</td>
			<td>Protocol</td>
			<td>SSL certificate</td>
		</tr>
	</thead>
	<tbody>
		''';
		
		for route in routes:
			
			if route["sslid"] in sslcerts:
				ssldisplay = sslcerts[route["sslid"]]["description"]
			else:
				ssldisplay = "-"
		
			if route["siteid"] in sites:
				sitedisplay = sites[route["siteid"]]["description"]
			else:
				sitedisplay = "[INVALID SITE]"
		
			html += '''
		<tr>
			<td>
				<a href="/route/''' + cgi.escape(urllib.quote(str(route["id"]))) + "\">" + cgi.escape(route["hostname"]) + '''</a>
			</td>
			<td>'''+cgi.escape(route["redirect_url"] if len(route["redirect_url"]) else sitedisplay)+'''</td>
			<td>'''+cgi.escape(str(len(self.srv.getInterfaceIDsForRoute(route["id"]) )))+'''</td>
			<td>'''+cgi.escape(SSLMODE_STRINGS[route["sslMode"]])+'''</td>
			<td>'''+cgi.escape(ssldisplay)+'''</td>
		</tr>
			''';
		
		html += '''
	</tbody>
	<tfoot>
		<tr>
			<td colspan="5">
				'''+str(len(routes))+''' route(s)
			</td>
		</tr>
	</tfoot>
</table>
		''';
		
		return self.templateBody(html);
	
	def routeRoute(self, id):
		if not self.isLoggedIn(): return self.routeIndex()
		
		site_options = dict({"0": "(no site)"}.items() + {x["siteid"] : x["description"] for x in self.srv.getSites()}.items())
		ssl_options  = dict({"0": "(no cert)"}.items() + {x["id"] : x["description"] for x in self.srv.getCerts()}.items())
		interface_list = {x["id"] : x["ipaddr"]+" ("+str(x["port_http"])+"/"+str(x["port_https"])+")" for x in self.srv.getInterfaces()}
		
		if id == "new":
			route = {
				"siteid":0,
				"hostname": "...",				
				"redirect_url":"",
				"sslMode":SSLMODE_HTTP_ONLY,
				"sslid":0
			}
		else:
			route = self.srv.getRoute(id)
			
			if route == False:
				return self.routeError("Invalid route selected.")
		
		html = '''
<strong>Routes &rsaquo; '''+cgi.escape(route["hostname"])+'''</strong><br>
<p>
</p>
<form method="POST" action="/post/updateRoute">
	<input type="hidden" name="id" value="'''+cgi.escape(id)+'''">
	<table>
		<thead>
			<tr>
				<td>&nbsp;</td>
				<td>&nbsp;</td>
			</tr>
		</thead>
		<tbody>
			<tr>
				<td>Hostname</td>
				<td><input type="text" name="hostname" value="'''+cgi.escape(route["hostname"])+'''"></td>
			</tr>
			<tr>
				<td>Redirect</td>
				<td><input type="text" name="redirect_url" value="'''+cgi.escape(route["redirect_url"])+'''"> (leave blank to connect to site)</td>
			</tr>
			<tr>
				<td>Site</td>
				<td>'''+HTML_select("siteid", route["siteid"], site_options)+'''</td>
			</tr>
			<tr>
				<td>Protocol</td>
				<td>'''+HTML_select("sslMode", route["sslMode"], SSLMODE_STRINGS)+'''</td>
			</tr>
			<tr>
				<td>SSL Certificate</td>
				<td>'''+HTML_select("sslid", route["sslid"], ssl_options)+'''</td>
			</tr>
			<tr>
				<td>Interfaces</td>
				<td>'''+HTML_multi('_interfaces', self.srv.getInterfaceIDsForRoute(route["id"]), interface_list)+'''</td>
			</tr>
		</tbody>
		<tfoot>
			<tr>
				<td>&nbsp;</td>
				<td><input type="submit" value="Save &raquo;"></td>
			</tr>
		</tfoot>
	</table>
</form>
		''';
		
		if id != "new":
			html += '''
<form method="POST" action="/post/updateRoute">
	<input type="hidden" name="id" value="'''+cgi.escape(id)+'''">
	<input type="hidden" name="_delete" value="1">
	<input type="submit" value="Delete &raquo;">
</form>
		''';
		
		return self.templateBody(html)
			
	def formatTimestamp(self, timestamp):
		return time.strftime("%Y-%m-%d", time.gmtime(int(timestamp)))
	
	def routeSSLCerts(self):
		if not self.isLoggedIn(): return self.routeIndex()
		
		sslcerts = self.srv.getCerts()
		html = '''
<strong>SSL Certificates</strong>
<p>
	<a href="/sslcert/new">Create new &raquo;</a>
</p>
<table>
	<thead>
		<tr>
			<td>Description</td>
			<td>Expiry</td>
		</tr>
	</thead>
	<tbody>
		''';
		
		for cert in sslcerts:
			html += '''
		<tr>
			<td>
				<a href="/sslcert/''' + cgi.escape(urllib.quote(str(cert["id"]))) + "\">" + cgi.escape(cert["description"]) + '''</a>
			</td>
			<td>'''+cgi.escape(self.formatTimestamp(cert["cache_expiry"]))+'''</td>
		</tr>
			''';
		
		html += '''
	</tbody>
	<tfoot>
		<tr>
			<td colspan="3">
				'''+str(len(sslcerts))+''' certificates(s)
			</td>
		</tr>
	</tfoot>
</table>
		''';
		
		return self.templateBody(html);
	
	def routeSSLCert(self, id):
		if not self.isLoggedIn(): return self.routeIndex()
		
		if id == "new":
			cert = {
				"id": "new",
				"description": "New SSL Certificate",
				"ssl_crt": "",
				"ssl_chain": "",
				"ssl_key": "",
				"cache_expiry": 0
			}
		else:
			cert = self.srv.getCert(id)
			if cert == False:
				return self.routeError("Invalid certificate selected.")
		
		html = '''
<strong>SSL Certificates &rsaquo; '''+cgi.escape(cert["description"])+'''</strong><br>
<p>
</p>
<form method="POST" action="/post/updateCert">
	<input type="hidden" name="id" value="'''+cgi.escape(id)+'''">
	<table>
		<thead>
			<tr>
				<td>&nbsp;</td>
				<td>&nbsp;</td>
			</tr>
		</thead>
		<tbody>
			<tr>
				<td>Description</td>
				<td>
					<input type="text" name="description" value="'''+cgi.escape(cert["description"])+'''">
				</td>
			</tr>
			<tr>
				<td>Certificate</td>
				<td><textarea name="ssl_crt" placeholder="-----BEGIN CERTFICIATE-----">'''+cgi.escape(str(cert["ssl_crt"]))+'''</textarea></td>
			</tr>
			<tr>
				<td>Chain</td>
				<td><textarea name="ssl_chain" placeholder="-----BEGIN CERTIFICATE-----">'''+cgi.escape(str(cert["ssl_chain"]))+'''</textarea></td>
			</tr>
			<tr>
				<td>Private key</td>
				<td><textarea name="ssl_key" placeholder="-----BEGIN PRIVATE KEY-----">'''+cgi.escape(str(cert["ssl_key"]))+'''</textarea></td>
			</tr>
			<tr>
				<td>Expires on</td>
				<td>'''+cgi.escape(self.formatTimestamp(cert["cache_expiry"]))+''' (cached)</td>
			</tr>
		</tbody>
		<tfoot>
			<tr>
				<td>&nbsp;</td>
				<td><input type="submit" value="Save &raquo;"></td>
			</tr>
		</tfoot>
	</table>
</form>
		''';
		
		if id != "new":
			html += '''
<form method="POST" action="/post/updateCert">
	<input type="hidden" name="id" value="'''+cgi.escape(id)+'''">
	<input type="hidden" name="_delete" value="1">
	<input type="submit" value="Delete &raquo;">
</form>
		''';
		
		return self.templateBody(html)
		
	def getInterfaceHTML(self):
		iplist = self.srv.getCurrentIPAddresses()
		return str(len(iplist)) + " interface(s) available on server: " + (", ".join(["<tt>"+cgi.escape(x)+"</tt>" for x in iplist]));
	
	def routeInterfaces(self):
		if not self.isLoggedIn(): return self.routeIndex()
		
		interfaces = self.srv.getInterfaces()
		html = '''
<strong>Interfaces</strong>
<p>
	<a href="/interface/new">Create new &raquo;</a>
</p>
<table>
	<thead>
		<tr>
			<td>IP Address</td>
			<td>HTTP Port</td>
			<td>HTTPS Port</td>
		</tr>
	</thead>
	<tbody>
		''';
		
		for iface in interfaces:
			html += '''
		<tr>
			<td>
				<a href="/interface/''' + cgi.escape(urllib.quote(str(iface["id"]))) + "\">" + cgi.escape(iface["ipaddr"]) + '''</a>
			</td>
			<td>'''+cgi.escape(str(iface["port_http"]))+'''</td>			
			<td>'''+cgi.escape(str(iface["port_https"]))+'''</td>
		</tr>
			''';
		
		html += '''
	</tbody>
	<tfoot>
		<tr>
			<td colspan="3">
				'''+str(len(interfaces))+''' interfaces(s)
			</td>
		</tr>
	</tfoot>
</table>
<p>
	''' + self.getInterfaceHTML() + '''
</p>
		''';
		
		return self.templateBody(html);
	
	def routeInterface(self, id):
		if not self.isLoggedIn(): return self.routeIndex()
		
		if id == "new":
			iface = {
				"id": "new",
				"ipaddr": "...",
				"port_http": 80,
				"port_https": 443,
				"addByDefault": 0
			}
		else:
			iface = self.srv.getInterface(id)
			if iface == False:
				return self.routeError("Invalid interface selected.")
		
		html = '''
<strong>Interface &rsaquo; '''+cgi.escape(iface["ipaddr"])+'''</strong><br>
<p>
</p>
<form method="POST" action="/post/updateInterface">
	<input type="hidden" name="id" value="'''+cgi.escape(id)+'''">
	<table>
		<thead>
			<tr>
				<td>&nbsp;</td>
				<td>&nbsp;</td>
			</tr>
		</thead>
		<tbody>
			<tr>
				<td>IP Address</td>
				<td>
					<input type="text" name="ipaddr" value="'''+cgi.escape(iface["ipaddr"])+'''">
					<span>
						Place IPv6 addresses in square brackets.
					</span>
				</td>
			</tr>
			<tr>
				<td>HTTP Port</td>
				<td><input type="number" min="1" max="65535" name="port_http" value="'''+cgi.escape(str(iface["port_http"]))+'''"></td>
			</tr>
			<tr>
				<td>HTTPS Port</td>
				<td><input type="number" min="1" max="65535" name="port_https" value="'''+cgi.escape(str(iface["port_https"]))+'''"></td>
			</tr>
			<tr>
				<td>Add by default to new routes</td>
				<td>'''+HTML_bool("addByDefault", iface["addByDefault"])+'''</td>
			</tr>
		</tbody>
		<tfoot>
			<tr>
				<td>&nbsp;</td>
				<td><input type="submit" value="Save &raquo;"></td>
			</tr>
		</tfoot>
	</table>
</form>
		''';
		
		if id != "new":
			html += '''
<form method="POST" action="/post/updateInterface">
	<input type="hidden" name="id" value="'''+cgi.escape(id)+'''">
	<input type="hidden" name="_delete" value="1">
	<input type="submit" value="Delete &raquo;">
</form>
		''';
		
		return self.templateBody(html)
		
	def routeGenConfig(self):
		if not self.isLoggedIn(): return self.routeIndex()
		
		config = self.srv.genConfig()
		return self.templateBody("<strong>Generated configuration</strong><br><p></p><textarea readonly>" + cgi.escape(config) + "</textarea>")
		
	def routeError(self, message):
		return self.templateBody("<strong>Message</strong><p>" + cgi.escape(message) + "</p><p><a href=\"/\">&laquo; Return</a></p>")
		
	def routeIndex(self):
		if self.isLoggedIn():
			return self.redirect("/sites")
		else:
			return self.redirect("/login")

# Entry point

if __name__ == "__main__":
	fmode = oct(stat.S_IMODE(os.lstat(CONFIG_FILE).st_mode))
	if fmode != "0600":
		print "[FATAL] Config file should have 0600 permissions, got %s" % fmode
		sys.exit(1)

	srv = AppServer(CONFIG_FILE)
	
	web = AppWebUI(srv)
	web.run()
	